from . import UbuntuPackageManager, RedHatPackageManager, SolarisPackageManager, RpmMixin
from infi import unittest

from infi.run_as_root import RootPermissions
from infi.pyutils.contexts import contextmanager

from infi import pkgmgr
#pylint: disable-all


def get_platform_name(): # pragma: no cover
    from platform import system
    name = system().lower().replace('-', '_')
    return name

def get_distribution(): # pragma: no cover
    """:returns: bunch with the following keys: distname, version, id
    """
    from munch import Munch
    from platform import linux_distribution
    distname, version, _id = linux_distribution()
    # distname in ['Red Hat Enterprise Linux Server', 'Ubuntu']
    distname = ''.join(distname.split()[:2]).lower()
    return Munch(distname=distname, version=version, id=_id.lower())

class TestOnUbuntu(unittest.TestCase):
    def _running_on_ubuntu(self):
        return get_platform_name() == "linux" and get_distribution().distname == "ubuntu"

    def setUp(self):
        super(TestOnUbuntu, self).setUp()
        self._should_skip()

    def _should_skip(self):
        if not self._running_on_ubuntu():
            raise self.skipTest("This test runs only on ubuntu")
        if not RootPermissions().is_root():
            raise self.skipTest("This test must run with root permissions")

    def test_sg3_utils(self):
        from infi.execute import execute
        execute('apt-get update'.split())
        self._check_package("sg3-utils", "/usr/bin/sg_inq")

    def _check_package(self, package_name, executable_name):
        pkgmgr = UbuntuPackageManager()
        is_installed_before = self._is_package_seems_to_be_installed(package_name, executable_name)
        self.assertEqual(pkgmgr.is_package_installed(package_name), is_installed_before)
        # Do the opposite
        pkgmgr.install_package(package_name) if not is_installed_before else pkgmgr.remove_package(package_name)
        self.assertNotEqual(pkgmgr.is_package_installed(package_name), is_installed_before)

    def _is_package_seems_to_be_installed(self, package_name, executable_name):
        from os.path import exists
        return exists(executable_name)

    def test_check_unknown_package(self):
        pkgmgr = UbuntuPackageManager()
        self.assertFalse(pkgmgr.is_package_installed('blablabla9988ok'))


class TestOnRedHat(unittest.TestCase):
    def _running_on_ubuntu(self):
        return get_platform_name() == "linux" and get_distribution().distname == "redhat"

    def setUp(self):
        super(TestOnRedHat, self).setUp()
        self._should_skip()

    def _should_skip(self):
        if not self._running_on_ubuntu():
            raise self.skipTest("This test runs only on ubuntu")
        if not RootPermissions().is_root():
            raise self.skipTest("This test must run with root permissions")

    def test_sg3_utils(self):
        self._check_package("sg3_utils", "/usr/bin/sg_inq")

    def _check_package(self, package_name, executable_name):
        pkgmgr = RedHatPackageManager()
        is_installed_before = self._is_package_seems_to_be_installed(package_name, executable_name)
        self.assertEqual(pkgmgr.is_package_installed(package_name), is_installed_before)
        # Do the opposite
        pkgmgr.install_package(package_name) if not is_installed_before else pkgmgr.remove_package(package_name)
        self.assertNotEqual(pkgmgr.is_package_installed(package_name), is_installed_before)

    def _is_package_seems_to_be_installed(self, package_name, executable_name):
        from os.path import exists
        return exists(executable_name)

from mock import patch

class Output(object):
    def __init__(self, returncode=0, stdout='', stderr=''):
        super(Output, self).__init__()
        self._returncode = returncode
        self._stdout = stdout
        self._stderr = stderr

    def get_stdout(self):
        return self._stdout

    def get_stderr(self):
        return self._stderr

    def get_returncode(self):
        return self._returncode

    def wait(self, timeout=None):
        pass

class TestUbuntuMock(TestOnUbuntu):
    def _should_skip(self):
        pass

    def _dpkg_query_s(self):
        from textwrap import dedent
        if self._installed:
            return Output(stdout=dedent("""
                                        Package: sg3-utils
                                        Status: installed ok installed
                                        Priority: optional
                                        Version: 1.30-1
                                        Section: admin
                                        """))
        else:
            return Output(stdout=dedent("""
                                        dpkg-query: package sg3-utils is not installed and no information is available
                                        Use dpkg --info (= dpkg-deb --info) to examine archive files,
                                        and dpkg --contents (= dpkg-deb --contents) to list their contents.
                                        """), returncode=1)

    def _dpkg_query_l(self):
        from textwrap import dedent
        return Output(stdout=dedent("""
                                    Desired=Unknown/Install/Remove/Purge/Hold
                                    | Status=Not/Inst/Conf-files/Unpacked/halF-conf/Half-inst/trig-aWait/Trig-pend
                                    |/ Err?=(none)/Reinst-required (Status,Err: uppercase=bad)
                                    ||/ Name                        Version            Architecture       Description
                                    +++-===========================-==================-==================-===========================================================
                                    {}  sg3-utils                   1.30-1             i386               utilities for devices using the SCSI command set
                                    """.format("ii" if self._installed else "un")))

    def _apt_get_install(self):
        self._installed = True
        return Output()

    def _apt_get_update(self):
        return Output()

    @contextmanager
    def _apply_patches(self):
        with patch("infi.execute.execute") as execute:
            def side_effect(*args, **kwargs):
                command = args[0]
                if "dpkg-query" in command:
                    if "-s" in command:
                        return self._dpkg_query_s()
                    if "-l" in command:
                        return self._dpkg_query_l()
                elif "apt-get install" in ' '.join(command):
                    return self._apt_get_install()
                elif "apt-get update" in ' '.join(command):
                    return self._apt_get_update()
                raise NotImplementedError()
            execute.side_effect = side_effect
            yield

    def test_sg3_utils(self):
        with self._apply_patches():
            super(TestUbuntuMock, self).test_sg3_utils()

    def test_check_unknown_package(self):
        with self._apply_patches():
            super(TestUbuntuMock, self).test_check_unknown_package()

    def setUp(self):
        self._installed = False

    def _is_package_seems_to_be_installed(self, package_name, executable_name):
        return self._installed

class TestRedHatMock(TestOnRedHat):
    def _should_skip(self):
        pass

    def _rpm_query(self):
        from textwrap import dedent
        return Output(stdout='sg3_utils-1.25-5.el5' if self._installed else 'package sg3_utils is not installed',
                      returncode=0 if self._installed else 1)

    def _yum_install(self):
        self._installed = True
        return Output()

    @contextmanager
    def _apply_patches(self):
        with patch("infi.execute.execute") as execute:
            def side_effect(*args, **kwargs):
                command = args[0]
                if "-q" in command:
                    return self._rpm_query()
                elif "install" in command:
                    return self._yum_install()
                raise NotImplementedError()
            execute.side_effect = side_effect
            yield

    def test_sg3_utils(self):
        with self._apply_patches():
            super(TestRedHatMock, self).test_sg3_utils()
        pass

    def setUp(self):
        self._installed = False

    def _is_package_seems_to_be_installed(self, package_name, executable_name):
        return self._installed

class test_package_versioning(unittest.TestCase):

    Solaris_v1 = """   VERSION:  6.0.100.000,REV=08.01.2012.09.00"""
    Solaris_v2 = """   VERSION:  5.14.2.5"""
    Ubuntu_v1 = """Version: 0.4.9-3ubuntu7.2"""
    Ubuntu_v2 = """Version: 1:1.2.8.dfsg-1ubuntu1"""
    rpm_v1 = """4.8-7.el7"""
    rpm_v2 = """18.168.6.1-34.el7"""
    def test_solaris_versioning_v1(self):
        with patch.object(pkgmgr , 'execute_command') as patched:
            patched().get_stdout.return_value = self.Solaris_v1
            patched().get_returncode.return_value = 0
            result = SolarisPackageManager().get_installed_version(self.Solaris_v1)
            self.assertEqual(result,{'version':'6.0.100.000', 'revision':'08.01.2012.09.00'})

    def test_solaris_versioning_v2(self):
        with patch.object(pkgmgr , 'execute_command') as patched:
            patched().get_stdout.return_value = self.Solaris_v2
            patched().get_returncode.return_value = 0
            result = SolarisPackageManager().get_installed_version(self.Solaris_v2)
            self.assertEqual(result,{'version':'5.14.2.5'})

    def test_ubuntu_versioning_v1(self):
        with patch.object(pkgmgr , 'execute_command') as patched:
            patched().get_stdout.return_value = self.Ubuntu_v1
            patched().get_returncode.return_value = 0
            result = UbuntuPackageManager().get_installed_version(self.Ubuntu_v1)
            self.assertEqual(result,{'version':'0.4.9-3ubuntu7.2'})

    def test_ubuntu_versioning_v2(self):
        with patch.object(pkgmgr , 'execute_command') as patched:
            patched().get_stdout.return_value = self.Ubuntu_v2
            patched().get_returncode.return_value = 0
            result = UbuntuPackageManager().get_installed_version(self.Ubuntu_v2)
            self.assertEqual(result,{'version':'1:1.2.8.dfsg-1ubuntu1'})

    def test_rpm_versioning_v1(self):
        with patch.object(pkgmgr , 'execute_command') as patched:
            patched().get_stdout.return_value = self.rpm_v1
            patched().get_returncode.return_value = 0
            result = RpmMixin().get_installed_version(self.rpm_v1)
            self.assertEqual(result,{'version':'4.8-7.el7'})

    def test_rpm_versioning_v2(self):
        with patch.object(pkgmgr , 'execute_command') as patched:
            patched().get_stdout.return_value = self.rpm_v2
            patched().get_returncode.return_value = 0
            result = RpmMixin().get_installed_version(self.rpm_v2)
            self.assertEqual(result,{'version':'18.168.6.1-34.el7'})

class GeneralTest(unittest.TestCase):
    def _is_solaris(self):
        from infi.os_info import get_platform_string
        return get_platform_string().split('-')[0] == 'solaris'

    def test_get_package_manager(self):
        package_manager = pkgmgr.get_package_manager()
        package_to_check = 'python'
        if self._is_solaris():
            package_to_check = 'CSW' + package_to_check
        self.assertTrue(package_manager.is_package_installed(package_to_check))



__import__("pkg_resources").declare_namespace(__name__)

from infi.pyutils.lazy import cached_method, cached_function, clear_cache

import logging # pylint: disable=W0403
logger = logging.getLogger()

WAIT_TIME = 120
QUERY_TIME = WAIT_TIME
INSTALL_TIME = 300

def execute_command(cmd, check_returncode=True, timeout=WAIT_TIME): # pragma: no cover
    from infi.execute import execute
    from os import environ
    logger.info("executing {}".format(cmd))
    env = environ.copy()
    env.pop('PYTHONPATH', 1)
    env['LC_ALL'] = 'en_US.UTF-8'
    process = execute(cmd, env=env)
    process.wait(WAIT_TIME)
    logger.info("execution returned {}".format(process.get_returncode()))
    logger.debug("stdout: {}".format(process.get_stdout()))
    logger.debug("stderr: {}".format(process.get_stderr()))
    if check_returncode and process.get_returncode() != 0:
        raise RuntimeError("execution of {} failed. see log file for more details".format(cmd))
    return process


class PackageManager(object): # pylint: disable=R0922
    def install_package(self, package_name):
        raise NotImplementedError() # pragma: no cover

    def is_package_installed(self, package_name):
        raise NotImplementedError() # pragma: no cover

    def remove_package(self, package_name):
        raise NotImplementedError() # pragma: no cover

    def get_installed_version(self, package_name):
        raise NotImplementedError() # pragma: nocover


class RpmMixin(object):
    def is_package_installed(self, package_name):
        cmd = "rpm -q {}".format(package_name).split()
        info = execute_command(cmd, timeout=QUERY_TIME, check_returncode=False)
        if info.get_returncode() == 0 and 'not installed' not in info.get_stdout():
            return True
        if info.get_returncode() == 1 and 'package {} is not installed'.format(package_name) in info.get_stdout():
            return False
        raise RuntimeError("rpm -q returned unexpected results, see the log")

    def get_installed_version(self, package_name):
        cmd = "rpm -q {rpm_name} --queryformat=%{{version}}-%{{release}}".format(rpm_name=package_name).split()
        info = execute_command(cmd, timeout=QUERY_TIME, check_returncode=False)
        if info.get_returncode() == 0 and 'not installed' not in info.get_stdout():
            return {'version':info.get_stdout().strip()}
        raise RuntimeError("Couldn't get package version")


class UbuntuPackageManager(PackageManager):
    def install_package(self, package_name, specific_version=None):
        cmd = "apt-get install -y {}".format(package_name)
        if specific_version:
            if self.get_installed_version(package_name)['version'] != specific_version:
                self.remove_package(package_name)
            cmd += "={} --force-yes".format(specific_version)
        execute_command(cmd.split(), timeout=INSTALL_TIME)

    def is_package_installed(self, package_name):
        cmd = "dpkg-query -l {}".format(package_name).split()
        dpkg_query = execute_command(cmd, check_returncode=False, timeout=QUERY_TIME)
        if dpkg_query.get_returncode() != 0:
            return False
        return self._extract_state_from_dpkg_query_output(dpkg_query.get_stdout()) == "ii"

    def _extract_state_from_dpkg_query_output(self, string):
        import re
        pattern = "^(?P<state>ii)\s+"
        match = re.search(pattern, string, re.MULTILINE)
        if match is None:
            return ''
        return match.groupdict()['state']

    def _extract_version_from_dpkg_query(self, dpkg_query):
        if dpkg_query.get_returncode() != 0:
            return ''
        else:
            import re
            pattern = "^Version:\s+(?P<version>[a-zA-Z0-9\.\-\_\:]+)$"
            match = re.search(pattern, dpkg_query.get_stdout(), re.MULTILINE)
            if match is None:
                return ''
            return match.groupdict()['version']

    def remove_package(self, package_name):
        cmd = "apt-get remove -y {}".format(package_name).split()
        execute_command(cmd, timeout=INSTALL_TIME)

    def get_installed_version(self, package_name):
        cmd = "dpkg-query -s {}".format(package_name).split()
        dpkg_query = execute_command(cmd, check_returncode=False, timeout=QUERY_TIME)
        return {'version':self._extract_version_from_dpkg_query(dpkg_query)}

class RedHatPackageManager(RpmMixin, PackageManager):
    def install_package(self, package_name):
        cmd = "yum install -y {}".format(package_name).split()
        execute_command(cmd, timeout=INSTALL_TIME)

    def remove_package(self, package_name):
        cmd = "yum remove -y {}".format(package_name).split()
        execute_command(cmd, timeout=INSTALL_TIME)


class SusePackageManager(RpmMixin, PackageManager):
    def install_package(self, package_name):
        cmd = "zypper --non-interactive --no-gpg-checks install --auto-agree-with-licenses {}".format(package_name).split()
        execute_command(cmd, timeout=INSTALL_TIME)

    def upgrade_package(self, package_name):
        cmd = "zypper --non-interactive --no-gpg-checks update --auto-agree-with-licenses {}".format(package_name).split()
        execute_command(cmd, timeout=INSTALL_TIME)

    def remove_package(self, package_name):
        cmd = "zypper --non-interactive --no-gpg-checks remove {}".format(package_name).split()
        execute_command(cmd, timeout=INSTALL_TIME)


class SolarisPackageManager(PackageManager):
    def install_package(self, package_name):
        # There is no unified package repository for solaris
        raise NotImplementedError

    def is_package_installed(self, package_name):
        cmd = "pkginfo {}".format(package_name).split()
        info = execute_command(cmd, timeout=QUERY_TIME, check_returncode=False)
        if info.get_returncode() == 0 and 'not found' not in info.get_stderr():
            return True
        elif info.get_returncode() == 1 and 'ERROR: information for "{}" was not found'.format(package_name) in info.get_stderr():
            return False
        else:
            raise RuntimeError("pkginfo returned unexpected results, see the log")

    def _extract_version_from_pkginfo_output(self, string):
        import re
        pattern = '[\ ]*VERSION:  (?P<version>[0-9\.]+)(\,REV\=)(?P<revision>[0-9\.\_\-]+)'
        match = re.search(pattern, string, re.MULTILINE)
        if match is None:
            pattern = '[\ ]*VERSION:  (?P<version>[0-9\.]+)'
            match = re.search(pattern, string, re.MULTILINE)
            if match is None:
                return ''
        return match.groupdict()

    def get_installed_version(self, package_name):
        """return dict of version and revision ( if exsist ) per pkg"""
        cmd = "pkginfo -l {}".format(package_name).split()
        pkginfo = execute_command(cmd, timeout=QUERY_TIME)
        return self._extract_version_from_pkginfo_output(pkginfo.get_stdout())

    def remove_package(self, package_name):
        raise NotImplementedError


def get_package_manager():
    from infi.os_info import get_platform_string
    platform_name = get_platform_string().split("-")[0]
    if platform_name == 'linux':
        # get distribution
        platform_name = get_platform_string().split("-")[1]
    pkgmgr_dict = {
        'ubuntu': UbuntuPackageManager,
        'redhat': RedHatPackageManager,
        'centos': RedHatPackageManager,
        'suse': SusePackageManager,
        'solaris': SolarisPackageManager
    }
    if platform_name in pkgmgr_dict:
        return pkgmgr_dict.get(platform_name)()
    else:
        raise RuntimeError("Package Manager is not implemented for {}".format(platform_name))
